unit xMemory;

interface

uses Windows, SysUtils, TlHelp32;

type
  tRange = record
    Start, Size: DWORD;
  end;
  TRangeList = array of tRange;
  
function MapGlobalData(const MapName: string; Size: Integer; var Ptr: Pointer): THandle;
procedure ReleaseGlobalData(var Handle: THandle; var Ptr: Pointer);

function IsGlobalDataExistent(const MapName: string): Boolean;

function GetProcessHeapList(SizeConstraint: DWORD = 0): TRangeList;

implementation

function MapGlobalData(const MapName: string; Size: Integer; var Ptr: Pointer): THandle;
begin
  Result := CreateFileMapping($FFFFFFFF, nil, PAGE_READWRITE, 0, Size, PChar(MapName));
  
  if Result = 0 then
    if GetLastError = ERROR_ALREADY_EXISTS then
    begin
      Result := OpenFileMapping(FILE_MAP_ALL_ACCESS, False, PChar(MapName));
      if Result = 0 then Exit;
    end else Exit;
  
  Ptr := MapViewOfFile(Result, FILE_MAP_ALL_ACCESS, 0, 0, 0);
  if Ptr = nil then
  begin
    CloseHandle(Result);
    Result := 0;
  end;
  
  // o, bo ZeroMemory | ?????????
end;

procedure ReleaseGlobalData(var Handle: THandle; var Ptr: Pointer);
begin
  if Assigned(Ptr) then
  begin
    UnmapViewOfFile(Ptr);
    Ptr := nil;
  end;

  if Handle <> 0 then
  begin
    CloseHandle(Handle);
    Handle := 0;
  end;
end;

function IsGlobalDataExistent(const MapName: string): Boolean;
var
  hMap: THandle;
begin
  hMap := OpenFileMapping(FILE_MAP_ALL_ACCESS, False, PChar(MapName));
  Result := hMap <> 0;
  if Result then CloseHandle(hMap);
end;

function GetProcessHeapList(SizeConstraint: DWORD = 0): TRangeList;
  
  procedure AddRange(RangeStart, RangeSize: DWORD);
  begin
    if (SizeConstraint <> 0) and (RangeSIze <> SizeConstraint) then Exit;

    SetLength(Result, Length(Result) + 1);
    with Result[Length(Result) - 1] do
    begin
      Start := RangeStart;
      Size := RangeSize;
    end;
  end;

  procedure GetProcessHeapList_NT;
  var
    HeapCount   : Integer;          
    Heaps, PHeap: PHandle;          
    I           : Integer;          
    HeapEntry   : TProcessHeapEntry;
    bResult     : Boolean;          
  begin
    HeapCount := GetProcessHeaps(0, Heaps^);
    if HeapCount = 0 then RaiseLastWin32Error;
    
    GetMem(Heaps, HeapCount * SizeOf(PHandle));
    try
      GetProcessHeaps(HeapCount, Heaps^);
      PHeap := Heaps;
      for I := 0 to HeapCount - 1 do
      begin
        FillChar(HeapEntry, SizeOf(HeapEntry), 0);
        bResult := HeapWalk(PHeap^, HeapEntry);
        while bResult do
        begin
          with HeapEntry do
            AddRange(DWORD(lpData), cbData);
          bResult := HeapWalk(PHeap^, HeapEntry);
        end;
        
        Inc(PHeap);
      end;
      
    finally
      FreeMem(Heaps);
    end;
  end;
  
  procedure GetProcessHeapList_95;
  var
    hSnapshot         : THandle;     
    bResult1, bResult2: Boolean;     
    HeapList          : THEAPLIST32; 
    HeapEntry         : THEAPENTRY32;
  begin
    hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPHEAPLIST, 0);
    if hSnapshot = 0 then
      RaiseLastWin32Error;
    
    HeapList.dwSize := SizeOf(HeapList);
    bResult1 := Heap32ListFirst(hSnapshot, HeapList);
    while bResult1 do
    begin
      HeapEntry.dwSize := SizeOf(HeapEntry);
      bResult2 := Heap32First(HeapEntry, HeapList.th32ProcessID, HeapList.th32HeapID);
      while bResult2 do
      begin
        with HeapEntry do
          AddRange(dwAddress, dwBlockSize);
        bResult2 := Heap32Next(HeapEntry);
      end;
      bResult1 := Heap32ListNext(hSnapshot, HeapList);
    end;
    
    CloseHandle(hSnapshot);
  end;
  
begin
  if Win32Platform = VER_PLATFORM_WIN32_NT then
    GetProcessHeapList_NT
  else
    GetProcessHeapList_95;
end;

end.

