unit xWindows;

interface

uses Windows, SysUtils, Controls, Forms;

const
  DEFAULT_FOCUSRECT_WIDTH = 3;
  
function GetWindowCaption(Wnd: THandle): string;
function GetWindowClassName(Wnd: THandle): string;

function FindControl(Handle: THandle; ProcessId: DWORD = 0): TWinControl;
function IsDelphiHandle(Handle: THandle; ProcessId: DWORD = 0): Boolean;

function FindTopmostWindow(ParentWnd: THandle; PT: TPoint): THandle;

procedure MyDrawFocusRect(DC: HDC; R: TRect; Width: Integer = DEFAULT_FOCUSRECT_WIDTH);
procedure DrawWindowFocusRect(Wnd: THandle; Width: Integer = DEFAULT_FOCUSRECT_WIDTH);
procedure DrawControlFocusRect(Control: TControl; Width: Integer = DEFAULT_FOCUSRECT_WIDTH);

function SearchChildByClassName(Window: THandle; WindwClassName: string): THandle;

implementation

uses xKernel;

var
  WindowAtom : TAtom;               
  ControlAtom: TAtom;               
  
  AtomText   : array[0..31] of Char;
  
  { Find a TWinControl given a window handle }
  
function FindControl(Handle: THandle; ProcessId: DWORD = 0): TWinControl;
begin
  //  DebugStrFmt('Find Control Handle = %x',[Handle]);
  if Handle <> 0 then
  begin
    if ProcessId = 0 then
      GetWindowThreadProcessId(Handle, @ProcessId);
    WindowAtom := GlobalAddAtom(StrFmt(AtomText, 'Delphi%.8X',[ProcessId]));
    Result := Pointer(GetProp(Handle, MakeIntAtom(WindowAtom)));
    //    DebugStrFmt('Find Control PID = %x Result = %p',[ProcessID, Pointer(Result)]);
  end else
    Result := nil;
end;

function IsDelphiHandle(Handle: THandle; ProcessId: DWORD = 0): Boolean;
begin
  Result := FindControl(Handle, ProcessId) <> nil;
end;

function FindTopmostWindow(ParentWnd: THandle; PT: TPoint): THandle;
var
  R: TRect;
begin
  Result := GetWindow(ParentWnd, GW_CHILD); // oĤ@Ӥl
  
  while Result <> 0 do
  begin
    GetWindowRect(Result, R); // oxΰϰ
    
    if IsWindowVisible(Result) and PtInRect(R, PT) then
    begin
      if Result = Application.MainForm.Handle then // { Main Form
      begin
        Result := 0;
        Exit;
      end;
      Result := FindTopmostWindow(Result, PT); // , jIsol
      Exit;
    end;
    Result := GetWindow(Result, GW_HWNDNEXT); // oU@ siblings 
  end;
  
  Result := ParentWnd; // 䤣, NOڭ̭n䪺
end;

{
  ϥ ChildWindowFromPoint ӴM, s non-visual |

function FindTopmostWindow(ParentWnd: THandle; Pt: TPoint): THandle;
begin
  repeat
    Result := ChildWindowFromPoint(ParentWnd, Pt);
    if Result = ParentWnd then break;

    if Result = Application.MainForm.Handle then // YO MainForm , Ǧ^ 0
    begin
      Result := 0;
      Exit;
    end;

    MapWindowPoints(ParentWnd, Result, Pt, 1);
    ParentWnd := Result;
  until false;
end;
}
procedure MyDrawFocusRect(DC: HDC; R: TRect; Width: Integer = DEFAULT_FOCUSRECT_WIDTH);
var
  I: Integer;
begin
  InflateRect(R, - Width, - Width);
  DrawFocusRect(DC, R);
  for I := 1 to Width do
  begin
    InflateRect(R, 1, 1);
    DrawFocusRect(DC, R);
  end;
end;

procedure DrawWindowFocusRect(Wnd: THandle; Width: Integer = DEFAULT_FOCUSRECT_WIDTH);
var
  DC: HDC;  
  R : TRect;
begin
  DC := GetWindowDC(Wnd);
  GetWindowRect(Wnd, R);
  MapWindowPoints(HWND_DESKTOP, Wnd, R, 2);
  OffsetRect(R, - R.Left, - R.Top);
  MyDrawFocusRect(DC, R, Width);
  ReleaseDC(Wnd, DC);
end;

procedure DrawControlFocusRect(Control: TControl; Width: Integer = DEFAULT_FOCUSRECT_WIDTH);
var
  Wnd: THandle; 
  DC : HDC;  
  R  : TRect;
begin
  Wnd := Control.Parent.Handle;
  DC := GetDC(Wnd);
  R := Control.BoundsRect;
  MyDrawFocusRect(DC, R, Width);
  ReleaseDC(Wnd, DC);
end;

function GetWindowCaption(Wnd: THandle): string;
var
  Buf: array[0..255] of Char;
begin
  GetWindowText(Wnd, Buf, SizeOf(Buf));
  Result := StrPas(Buf);
end;

function GetWindowClassName(Wnd: THandle): string;
var
  Buf: array[0..255] of Char;
begin
  GetClassName(Wnd, Buf, SizeOf(Buf));
  Result := StrPas(Buf);
end;

function SearchChildByClassName(Window: THandle; WindwClassName: string): THandle;
var
  ResultWnd: THandle;
  
  function SearchChild(Wnd: THandle): THandle; stdcall; forward;
    
    function EnumChildWindowProc(Wnd: THandle; LPARAM: LPARAM): Boolean; stdcall;
    var
      S: string;
    begin
      DebugStr('EnumChildWindowProc enters');
      
      S := GetWindowClassName(Wnd);
      DebugStrFmt('EnumChildWindowProc: %x %s',[Wnd, S]);
      
      if CompareText(S, WindwClassName) = 0 then
      begin
        ResultWnd := Wnd;
        DebugStrFmt('Found ResultWnd window: %x',[Wnd]);
        Result := False;
        Exit;
      end;

      DebugStr('Before SearchChild in EnumChildWindowProc');
      SearchChild(Wnd);
      DebugStr('After SearchChild in EnumChildWindowProc');
      
      Result := (ResultWnd = 0);
    end;
    
    function SearchChild(Wnd: THandle): THandle; stdcall;
    begin
{      if not IsWindow(Wnd) then Exit;
      if not IsWindowVisible(Wnd) then Exit;
      if GetWindow(Wnd, GW_CHILD) = 0 then Exit;
}      
      DebugStrFmt('SearchChild: %s',[GetWindowClassName(Wnd)]);
      EnumChildWindows(Wnd, @EnumChildWindowProc, 0);
      Result := ResultWnd;
    end;
    
  begin
    ResultWnd := 0;
    Result := SearchChild(Window);
  end;
  
end.

