//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <registry.hpp>
#include <mmsystem.h>
#include <extctrls.hpp>

#include "main.h"
#include "xFiles.hpp"
#include "xUtils.hpp"
#include "xGraphics.hpp"
#include "xRegistry.hpp"
#include "xDesktop.hpp"
#include "xFonts.hpp"
#include "xControls.hpp"
#include "about.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

AnsiString __fastcall TMainForm::ReadThemeItem(TIniFile* IniFile, AnsiString Section, AnsiString Ident)
{
  if (Ident == "") Ident = "DefaultValue";
  AnsiString value = IniFile->ReadString(Section, Ident, "");

  AnsiString S = FThemePath + ExtractFileName(value);
  if (FileExistsAfterTruncate(S)) value = S;

  return value;
}

void __fastcall TMainForm::ReadThemeFile(AnsiString FileName)
{
  DoBusy(true);
  try {
    FThemePath = PathWithSlash(ExtractFilePath(FileName), '\\');
    FThemeName = ExtractFileName(FileName);

    TIniFile* IniFile = new TIniFile(FileName);
    try {
      // read metrics
      AnsiString S;
      S = ReadThemeItem(IniFile, secMetrics, "IconMetrics");
      StringToBuffer(S, &FIM, sizeof(FIM));
      S = ReadThemeItem(IniFile, secMetrics, "NonClientMetrics");
      StringToBuffer(S, &FNM, sizeof(FNM));

      FMyFonts[mfIcon] = FIM.lfFont;
      FMyFonts[mfCaptionFont] = FNM.lfCaptionFont;
      FMyFonts[mfsmCaptionFont] = FNM.lfSmCaptionFont;
      FMyFonts[mfMenuFont] = FNM.lfMenuFont;
      FMyFonts[mfStatusFont] = FNM.lfStatusFont;
      FMyFonts[mfMessageFont] = FNM.lfMessageFont;

      // read my icons
      if (IniFile->SectionExists(secMyComputerKey))
        FMyIcons[mcMyComputer] = ReadThemeItem(IniFile, secMyComputerKey, ""); // for fool win98"s theme
      else
        FMyIcons[mcMyComputer] = ReadThemeItem(IniFile, secMyComputer, "");

      if (IniFile->SectionExists(secNeighborKey))
        FMyIcons[mcNeighbor] = ReadThemeItem(IniFile, secNeighborKey, "");
      else
        FMyIcons[mcNeighbor] = ReadThemeItem(IniFile, secNeighbor, "");

      if (IniFile->SectionExists(secRecycleBinKey)) {
        FMyIcons[mcRecycleBinFull] = ReadThemeItem(IniFile, secRecycleBinKey, "full");
        FMyIcons[mcRecycleBinEmpty] = ReadThemeItem(IniFile, secRecycleBinKey, "empty");
      } else {
        FMyIcons[mcRecycleBinFull] = ReadThemeItem(IniFile, secRecycleBin, "full");
        FMyIcons[mcRecycleBinEmpty] = ReadThemeItem(IniFile, secRecycleBin, "empty");
      }

      // read my colors
      for (TMyColor i = 0; i <= TMyColor_MAX; i = i + 1) {
        AnsiString S = ReadThemeItem(IniFile, secColors, MyColorID[i]);
        if (S != "")
          FMyColors[i] = RGBStringToColor(S);
        else {
          if (i == GradientActiveTitle)
            FMyColors[i] = FMyColors[ActiveTitle];
          else if (i == GradientInactiveTitle)
            FMyColors[i] = FMyColors[InactiveTitle];
          else
            FMyColors[i] = -1; // w]
        }
      }

      // read my cursors
      for (TMyCursor i = 0; i <= TMyCursor_MAX; i = i + 1)
        FMyCursors[i] = ReadThemeItem(IniFile, secCursors, MyCursorID[i]);

      if (FMyCursors[mcrArrow] != "") {
        Screen->Cursors[200] = LoadCursorFromFile(FMyCursors[mcrArrow].c_str());
        imgPreview->Cursor = 200;
      }

      // read my sounds
      for (TMySound i = 0; i <= TMySound_MAX; i = i + 1)
        FMySounds[i] = ReadThemeItem(IniFile, AnsiString("AppEvents\\Schemes\\Apps\\") + MySoundSect[i], "");

      // read others
      FWallPaper = ReadThemeItem(IniFile, secDesktop, "WallPaper");

      FWallPaperStyle = StrToIntDef(ReadThemeItem(IniFile, secDesktop, "WallPaperStyle"), 0);
      FTileWallPaper = StrToIntDef(ReadThemeItem(IniFile, secDesktop, "TileWallPaper"), 0);
      FPattern = ReadThemeItem(IniFile, secDesktop, "Pattern");

      FScreenSaver = ReadThemeItem(IniFile, secBoot, "SCRNSAVE.EXE");
    } __finally {
      delete IniFile;
    }

    LoadThemeComponents();
    DrawTheme(imgPreview->Canvas);

    Invalidate();
  } __finally {
    DoBusy(false);
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadDesc()
{
  lbxColors->Items->Clear();
  for (TMyColor i = 0; i <= TMyColor_MAX; i = i + 1)
    lbxColors->Items->Add(MyColorDesc[i]);

  lbxCursors->Items->Clear();
  for (TMyCursor i = 0; i <= TMyCursor_MAX; i = i + 1)
    lbxCursors->Items->Add(MyCursorDesc[i]);

  lbxSounds->Items->Clear();
  TRegIniFile* r = new TRegIniFile("AppEvents\\EventLabels");
  try {
    for (TMySound i = 0; i <= TMySound_MAX; i = i + 1)
      lbxSounds->Items->Add(r->ReadString(MySoundDescKey[i], "", MySoundDescKey[i]));
  } __finally {
    delete r;
  }

  lbxFonts->Items->Clear();
  for (TMyFont i = 0; i <= TMyFont_MAX; i = i + 1)
    lbxFonts->Items->Add(MyFontDesc[i]);

  lbxDesktop->Items->Clear();
  for (TMyIcon i = 0; i <= TMyIcon_MAX; i = i + 1)
    lbxDesktop->Items->Add(MyIconDesc[i]);

  for (TMyDesktop i = 0; i <= TMyDesktop_MAX; i = i + 1)
    lbxDesktop->Items->Add(MyDesktopDesc[i]);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadThemeComponents()
{
  for (TMyCursor i = 0; i <= TMyCursor_MAX; i = i + 1)
    FMyCursorIcons[i] = (HCURSOR)CursorToIcon((THandle)MyLoadCursor(i, FMyCursors[i]), true);

  for (TMyIcon i = 0; i <= TMyIcon_MAX; i = i + 1)
    FMyIconIcons[i] = (HICON)MyLoadIcon(FMyIcons[i], 0);

  FWallPaperBits->ReleaseHandle();

  if (FileExists(FWallPaper)) {
    typedef HBITMAP __stdcall (*TNViewLibLoadProc)(char* FileName, bool ShowProgress);
    TNViewLibLoadProc NViewLibLoadProc;

    HINSTANCE hLib;
    if ((hLib = LoadLibrary("NViewLib.dll")) != NULL) {

      NViewLibLoadProc = (TNViewLibLoadProc)GetProcAddress(hLib, "NViewLibLoad");
      if (NViewLibLoadProc)
        FWallPaperBits->Handle = NViewLibLoadProc(FWallPaper.c_str(), false);

      FreeLibrary(hLib);
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawTheme(TCanvas* ACanvas)
{
  DrawBackgound(ACanvas);
  DrawIcons(ACanvas);
  DrawWindows(ACanvas);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateCaption()
{
  if (FThemeName != "")
    Caption = Format("%s - %s%s", OPENARRAY(TVarRec,
      (Application->Title, FThemePath, FThemeName)));
  else
    Caption = Format("%s - ثe]w", OPENARRAY(TVarRec,
      (Application->Title)));
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ApplyTheme()
{
  DoBusy(true);
  try {
    TRegIniFile* r = new TRegIniFile("Control Panel");
    try {
      r->WriteString("Desktop", "WallPaperStyle", IntToStr(FWallPaperStyle));
      r->WriteString("Desktop", "TileWallPaper", IntToStr(FTileWallPaper));

      if (FWallPaper == "")
        SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, "", SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE);
      else if (ExtractFileExt(AnsiUpperCase(FWallPaper)) == ".BMP")
        SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, FWallPaper.c_str(), SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE);
      else {
        FWallPaperBits->SaveToFile(GetMyWallPaperFile());
        SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, GetMyWallPaperFile().c_str(), SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE);
      }

      SystemParametersInfo(SPI_SETDESKPATTERN, 0, FPattern.c_str(), SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE);
      r->WriteString("Desktop", "SCRNSAVE.EXE", FScreenSaver);

      // N cursors gJ registry
      for (TMyCursor i = 0; i <= TMyCursor_MAX; i = i + 1)
        r->WriteString("Cursors", MyCursorID[i], FMyCursors[i]);
    } __finally {
      delete r;
    }
    
    TRegistry* Registry = new TRegistry;
    try {
      // gJ Sounds
      for (TMySound i = 0; i <= TMySound_MAX; i = i + 1) {
        if (Registry->OpenKey(AnsiString("AppEvents\\Schemes\\Apps\\") + MySoundSect[i], true)) {
          Registry->WriteString("", FMySounds[i]);
          Registry->CloseKey();
        }
      }
        
      // gJ Icon
      Registry->RootKey = HKEY_CLASSES_ROOT;
      WriteKeyDefaultValue(Registry, secMyComputer, TruncateTrailIfNotDLL(FMyIcons[mcMyComputer]));
      WriteKeyDefaultValue(Registry, secNeighbor, TruncateTrailIfNotDLL(FMyIcons[mcNeighbor]));
        
      if (Registry->OpenKey(secRecycleBin, true)) {
        Registry->WriteString("Full", TruncateTrailIfNotDLL(FMyIcons[mcRecycleBinFull]));
        Registry->WriteString("Empty", TruncateTrailIfNotDLL(FMyIcons[mcRecycleBinEmpty]));
        Registry->WriteString("", TruncateTrailIfNotDLL(FMyIcons[mcRecycleBinEmpty]));
          
        Registry->CloseKey();
      }
    } __finally {
      delete Registry;
    }

    //  cursors
    for (TMyCursor i = 0; i <= TMyCursor_MAX; i = i + 1) {
      // win95/98  animated cursor  setsystemcursor D, |
      if (AnsiUpperCase(ExtractFileExt(FMyCursors[i])) == ".ANI" &&
        Win32Platform != VER_PLATFORM_WIN32_NT) continue;

      HCURSOR hc = MyLoadCursor(i, FMyCursors[i]);
      if (hc != 0) SetSystemCursor(hc, MyCursorNo[i]);
    }

    // Set colors
    int SaveColorNo[41];
    COLORREF SaveColors[41];

    int CC = 0; // count of colors
    for (TMyColor i = 0; i <= TMyColor_MAX; i = i + 1) {
      SaveColorNo[CC] = MyColorNo[i];
      SaveColors[CC] = FMyColors[i];
      CC++;
    }
    // ]wC
    SetSysColors(CC, SaveColorNo, SaveColors);

    // nxsb registry
    r = new TRegIniFile("Control Panel");
    try {
      for (TMyColor i = 0; i <= TMyColor_MAX; i = i + 1)
        r->WriteString("Colors", MyColorID[i], ColorToRGBString(FMyColors[i]));
    } __finally {
      delete r;
    }

    // Set Metrics
    FIM.cbSize = sizeof(FIM);
    SystemParametersInfo(SPI_SETICONMETRICS, sizeof(FIM), &FIM, SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE);

    FNM.cbSize = sizeof(FNM);
    SystemParametersInfo(SPI_SETNONCLIENTMETRICS, sizeof(FNM), &FNM, SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE);

    // jୱϥܧs
    RebuildIconCache();
  } __finally {
    DoBusy(false);
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadCurrentSettings()
{
  FThemeName = "";

  memset(&FIM, sizeof(FIM), 0);
  FIM.cbSize = sizeof(FIM);
  SystemParametersInfo(SPI_GETICONMETRICS, sizeof(FIM), &FIM, 0);

  memset(&FNM, sizeof(FNM), 0);
  FNM.cbSize = sizeof(FNM);
  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(FNM), &FNM, 0);
  
  FMyFonts[mfIcon] = FIM.lfFont;
  FMyFonts[mfCaptionFont] = FNM.lfCaptionFont;
  FMyFonts[mfsmCaptionFont] = FNM.lfSmCaptionFont;
  FMyFonts[mfMenuFont] = FNM.lfMenuFont;
  FMyFonts[mfStatusFont] = FNM.lfStatusFont;
  FMyFonts[mfMessageFont] = FNM.lfMessageFont;
  
  TRegistry* Registry = new TRegistry;
  try {
    FMyIcons[mcMyComputer] = "";
    FMyIcons[mcNeighbor] = "";
    FMyIcons[mcRecycleBinFull] = "";
    FMyIcons[mcRecycleBinEmpty] = "";

    for (int bLocalDesktopIcons = 0; bLocalDesktopIcons <= 1; bLocalDesktopIcons++) {
      if (bLocalDesktopIcons)
        Registry->RootKey = HKEY_LOCAL_MACHINE;
      else
        Registry->RootKey = HKEY_CURRENT_USER;
      
      // desktop icons ..
      if (FMyIcons[mcMyComputer] == "")
        FMyIcons[mcMyComputer] = ReadKeyDefaultValue(Registry, secMyComputerKey);
      if (FMyIcons[mcNeighbor] == "")
        FMyIcons[mcNeighbor] = ReadKeyDefaultValue(Registry, secNeighborKey);
      
      if (Registry->OpenKey(secRecycleBinKey, false)) {
        if (FMyIcons[mcRecycleBinFull] == "")
          FMyIcons[mcRecycleBinFull] = Registry->ReadString("Full");
        if (FMyIcons[mcRecycleBinEmpty] == "")
          FMyIcons[mcRecycleBinEmpty] = Registry->ReadString("Empty");
        Registry->CloseKey();
      }
    }

    Registry->RootKey = HKEY_CURRENT_USER;
    // sounds
    for (TMySound i = 0; i <= TMySound_MAX; i = i + 1)
      FMySounds[i] = ReadKeyDefaultValue(Registry, AnsiString("AppEvents\\Schemes\\Apps\\") + MySoundSect[i]);
  } __finally {
    delete Registry;
  }

  TRegIniFile* r = new TRegIniFile("Control Panel");
  try {
    // others
    FWallPaperStyle = StrToIntDef(r->ReadString("Desktop", "WallPaperStyle", "0"), 0);
    FTileWallPaper = StrToIntDef(r->ReadString("Desktop", "TitleWallPaper", "0"), 0);

    FWallPaper = r->ReadString("Desktop", "WallPaper", "");

    FScreenSaver = r->ReadString("Desktop", "SCRNSAVE.EXE", "");
    FPattern = r->ReadString("Desktop", "Pattern", "");

    //  cursors
    for (TMyCursor i = 0; i <= TMyCursor_MAX; i = i + 1)    
      FMyCursors[i] = r->ReadString("Cursors", MyCursorID[i], "");
  } __finally {
    delete r;
  }
  
  // Colors
  for (TMyColor i = 0; i <= TMyColor_MAX; i = i + 1)
    FMyColors[i] = GetSysColor(MyColorNo[i]);

  LoadThemeComponents();
  DrawTheme(imgPreview->Canvas);

  Invalidate();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuOpenClick(TObject *Sender)
{
  if (dlgOpen->Execute()) {
    ReadThemeFile(dlgOpen->FileName);
    UpdateCaption();
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DrawBackgound(TCanvas* ACanvas)
{
  // I
  ACanvas->Brush->Style = bsSolid;
  ACanvas->Brush->Color = FMyColors[Background];
  ACanvas->FillRect(ACanvas->ClipRect);

  if (FileExists(FWallPaper))
    ACanvas->StretchDraw(ACanvas->ClipRect, FWallPaperBits);
}

void __fastcall TMainForm::DrawIcons(TCanvas* ACanvas)
{
  const int IconX         = 75;
  const int IconY         = 40;

  const int offX          = 32 / 2; // icon size @wO 32
  const int offY          = 40;

  ACanvas->Brush->Style = bsSolid;
  ACanvas->Brush->Color = FMyColors[Background];
  ACanvas->Font->Color = CalcForeColor(FMyColors[Background]);
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfIcon]);

  SetTextAlign(ACanvas->Handle, TA_CENTER);
  for (TMyIcon i = (TMyIcon)0; i <= TMyIcon_MAX; i = i + 1) {
    HCURSOR H = (HCURSOR)MyLoadIcon(FMyIcons[i], 0);
    DrawIcon(ACanvas->Handle, IconX, IconY + FIM.iVertSpacing * i, H);
    DestroyIcon(H);

    ACanvas->TextOut(IconX + offX, IconY + FIM.iVertSpacing * i + offY, MyIconDesc[i]);

    int wid = ACanvas->TextWidth(MyIconDesc[i]);
    int hei = ACanvas->TextHeight(MyIconDesc[i]);
    TRect R = Rect(IconX + offX, IconY + FIM.iVertSpacing * i + offY, IconX + offX, IconY + FIM.iVertSpacing * i + offY);
    InflateRect(&R, wid / 2, hei / 2);
  }

  SetTextAlign(ACanvas->Handle, TA_LEFT);
}

void __fastcall TMainForm::DrawBorderIcon(TCanvas* ACanvas, TRect ARect, int Kind)
{
  int OldLeft, OldTop;

  TRect RT = ARect;
  Frame3D(ACanvas, ARect, FMyColors[ButtonHilight], FMyColors[ButtonDkShadow], 1);
  Frame3D(ACanvas, ARect, FMyColors[ButtonFace], FMyColors[ButtonShadow], 1);
  ACanvas->Brush->Color = FMyColors[ButtonFace];
  ACanvas->FillRect(ARect);

  int W = RT.Right - RT.Left;
  int H = RT.Bottom - RT.Top;
  int H1;

  // 1: ̤p 2: ̤j 3: 
  switch (Kind) {
    case 1:
      RT.Right = RT.Left + W * 63 / 100;
      RT.Left = RT.Left + W * 20 / 100;
      RT.Bottom = RT.Top + H * 85 / 100;
      RT.Top = RT.Top + H * 70 / 100;
      ACanvas->Brush->Color = clBlack;
      ACanvas->FillRect(RT);
      break;

    case 2:
      OldLeft = RT.Left;
      OldTop = RT.Top;

      RT.Right = RT.Left + W * 80 / 100;
      RT.Left = RT.Left + W * 18 / 100;
      RT.Bottom = RT.Top + H * 83 / 100;
      RT.Top = RT.Top + H * 18 / 100;
      ACanvas->Brush->Color = clBlack;
      ACanvas->FillRect(RT);

      RT.Left = OldLeft;
      RT.Top = OldTop;
      RT.Right = RT.Left + W * 74 / 100;
      RT.Left = RT.Left + W * 24 / 100;
      RT.Bottom = RT.Top + H * 77 / 100;
      RT.Top = RT.Top + H * 31 / 100;
      ACanvas->Brush->Color = FMyColors[ButtonFace];
      ACanvas->FillRect(RT);
      break;
          
    case 3:
      ACanvas->Pen->Color = clBlack;
      ACanvas->Pen->Width = W / 8;
      H1 = H * 60 / 100;
      RT.Left = RT.Left + W * 20 / 100;
      RT.Top = RT.Top + H * 16 / 100;
      RT.Right = RT.Left + H1;
      RT.Bottom = RT.Top + H1;

      ACanvas->MoveTo(RT.Left, RT.Top);
      ACanvas->LineTo(RT.Right, RT.Bottom);

      ACanvas->MoveTo(RT.Right, RT.Top);
      ACanvas->LineTo(RT.Left, RT.Bottom);
      break;
  }
}

void __fastcall TMainForm::DrawScrollButton(TCanvas* ACanvas, TRect ARect, int Kind)
{
  TRect RT = ARect;

  Frame3D(ACanvas, RT, FMyColors[ButtonFace], FMyColors[ButtonDkShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[ButtonHilight], FMyColors[ButtonShadow], 1);
  ACanvas->Brush->Color = FMyColors[ButtonFace];
  ACanvas->FillRect(RT);

  ACanvas->Brush->Color = FMyColors[ButtonDkShadow];
  int W = RT.Right - RT.Left;
  int H = RT.Bottom - RT.Top;

  // 0: up arrow 1: down arrow
  TPoint points[3];
  switch (Kind) {
    case 0:
      points[0] = Point(RT.Left + W * 0.16, RT.Top + H * 0.65);
      points[1] = Point(RT.Left + W * 0.74, RT.Top + H * 0.65);
      points[2] = Point(RT.Left + W * 0.46, RT.Top + H * 0.25);
      ACanvas->Polygon(points, 2);
      break;

    case 1:
      points[0] = Point(RT.Left + W * 0.16, RT.Top + H * 0.25);
      points[1] = Point(RT.Left + W * 0.74, RT.Top + H * 0.25);
      points[2] = Point(RT.Left + W * 0.46, RT.Top + H * 0.65);
      ACanvas->Polygon(points, 2);
      break;
  }
}

void FillGradientRect(TCanvas* ACanvas, TRect RT, TColor LeftColor, TColor RightColor)
{
  if (LeftColor == RightColor) {
    ACanvas->Brush->Color = LeftColor;
    ACanvas->FillRect(RT);
    return;
  }

  int W = RT.Right - RT.Left;
  Byte R1 = GetRValue(ColorToRGB(LeftColor));
  Byte G1 = GetGValue(ColorToRGB(LeftColor));
  Byte B1 = GetBValue(ColorToRGB(LeftColor));

  float RD = (GetRValue(ColorToRGB(RightColor)) - R1) / W;
  float GD = (GetGValue(ColorToRGB(RightColor)) - G1) / W;
  float BD = (GetBValue(ColorToRGB(RightColor)) - B1) / W;

  W = 0;
  for (int i = RT.Left; i < RT.Right; i++) {
    ACanvas->Pen->Color = RGB(R1 + RD * W, G1 + GD * W, B1 + BD * W);
    ACanvas->MoveTo(i, RT.Top);
    ACanvas->LineTo(i, RT.Bottom);
    W++;
  }
}

void __fastcall TMainForm::DrawWindows(TCanvas* ACanvas)
{
  TRect RT, oldRT;
  TRect rtInactive, rtActive, rtMemo, rtMsgBox;
  TPoint Pos;

  // NONCLIENTMETRICS GetSYSColor
  // eDu@
  RT = Rect(185, 145, 415, 330);
  rtInactive = RT;

  // Frame3D |۰ InflateRect...
  Frame3D(ACanvas, RT, FMyColors[ButtonFace], FMyColors[ButtonDkShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[ButtonHilight], FMyColors[ButtonShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[InactiveBorder], FMyColors[InactiveBorder], FNM.iBorderWidth);
      
  // border
  Brush->Color = FMyColors[ButtonFace];
  ACanvas->FillRect(RT);
      
  // caption
  InflateRect(&RT, - 1, - 1);
  //    FNM.iCaptionwidth = 30;
  //    FNM.iCaptionheight = 30;
  RT.Bottom = RT.Top + FNM.iCaptionHeight;
  FillGradientRect(ACanvas, RT, FMyColors[InactiveTitle], FMyColors[GradientInactiveTitle]);
      
  ACanvas->Brush->Style = bsClear;
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfCaptionFont]);
  ACanvas->Font->Color = FMyColors[InactiveTitleText];
      
  Pos.x = RT.Left + 2; // LOGFONT
  Pos.y = RT.Top + (RT.Bottom - RT.Top - ACanvas->TextHeight(sInactiveWindow)) / 2;
  ACanvas->TextRect(RT, Pos.x, Pos.y, sInactiveWindow);
  oldRT = RT; //  RT d_ӫݷ|n
  ACanvas->Brush->Style = bsSolid;

  // eees
  int btnSize = FNM.iCaptionWidth - 2;
  InflateRect(&RT, - 2, - 2);

  RT.Left = RT.Right - btnSize;
  DrawBorderIcon(ACanvas, RT, 3); // 

  OffsetRect(&RT, - (btnSize + 1), 0);
  DrawBorderIcon(ACanvas, RT, 2); // ̤j

  OffsetRect(&RT, - btnSize, 0);
  DrawBorderIcon(ACanvas, RT, 1); // ̤p

  // e Active Window
  // oɭ oldRT ODϥΤ Title
  RT = oldRT;
  RT.Left = RT.Left + 1;
  RT.Top = RT.Bottom + 1;
  RT.Right = rtInactive.Right + btnSize + 6 + FNM.iBorderWidth;
  RT.Bottom = rtInactive.Bottom + 4;
  rtActive = RT;

  // Frame3D |۰ InflateRect...
  Frame3D(ACanvas, RT, FMyColors[ButtonFace], FMyColors[ButtonDkShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[ButtonHilight], FMyColors[ButtonShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[ActiveBorder], FMyColors[ActiveBorder], FNM.iBorderWidth);
      
  // border
  ACanvas->Brush->Color = FMyColors[ButtonFace];
  ACanvas->FillRect(RT);
      
  // caption
  InflateRect(&RT, - 1, - 1);
  RT.Bottom = RT.Top + FNM.iCaptionHeight;
      
  FillGradientRect(ACanvas, RT, FMyColors[ActiveTitle], FMyColors[GradientActiveTitle]);
      
  ACanvas->Brush->Style = bsClear;
  ACanvas->Font->Color = FMyColors[TitleText];
  Pos.x = RT.Left + 2;
  Pos.y = RT.Top + (RT.Bottom - RT.Top - ACanvas->TextHeight(sActiveWindow)) / 2;
  ACanvas->TextRect(RT, Pos.x, Pos.y, sActiveWindow);
  ACanvas->Brush->Style = bsSolid;
      
  // eees
  btnSize = FNM.iCaptionWidth - 2;
  InflateRect(&RT, - 2, - 2);

  oldRT = RT; //  RT d_ӫݷ|n
  RT.Left = RT.Right - btnSize;
  DrawBorderIcon(ACanvas, RT, 3); // 

  OffsetRect(&RT, - (btnSize + 1), 0);
  DrawBorderIcon(ACanvas, RT, 2); // ̤j

  OffsetRect(&RT, - btnSize, 0);
  DrawBorderIcon(ACanvas, RT, 1); // ̤p

  // NONCLIENTMETRICS
  // oɭ oldRT OϥΤ Title
  RT = oldRT;
  OffsetRect(&RT, 0, FNM.iCaptionHeight);
  RT.Bottom = RT.Top + FNM.iMenuHeight;
  oldRT = RT;
      
  // \
  ACanvas->Brush->Color = FMyColors[MenuColor];
  ACanvas->FillRect(RT);
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfMenuFont]);
  ACanvas->Font->Color = FMyColors[MenuText];
      
  RT.Top = RT.Top - 1;
  Pos.x = RT.Left + 4; // LOGFONT
  Pos.y = RT.Top + (RT.Bottom - RT.Top -
    ACanvas->TextHeight(AnsiString(sNormal) + sDisabled + sSelected)) / 2;
  //    TextOut(Pos.x, Pos.y, "з");
  ACanvas->TextOut(Pos.x, Pos.y, sNormal);
      
  Pos.x = RT.Left + ACanvas->TextWidth(AnsiString(sNormal) + "12");
      
  ACanvas->Font->Color = FMyColors[ButtonHilight];
  ACanvas->TextOut(Pos.x + 1, Pos.y + 1, sDisabled); // kU
  ACanvas->Font->Color = FMyColors[GrayText];
  //    TextOut(Pos.x, Pos.y, "");
  ACanvas->Brush->Style = bsClear;
  ACanvas->TextOut(Pos.x, Pos.y, sDisabled);
  ACanvas->Brush->Style = bsSolid;
      
  Pos.x = Pos.x + ACanvas->TextWidth(AnsiString(sDisabled) + "123");
  ACanvas->Brush->Color = FMyColors[Hilight];
  ACanvas->Font->Color = FMyColors[HilightText];
  RT.Left = Pos.x - ACanvas->TextWidth("1");
  RT.Right = Pos.x + ACanvas->TextWidth(sSelected) + ACanvas->TextWidth("1");
  //    TextRect(RT, Pos.x, Pos.y, "");
  RT.Bottom = RT.Bottom - 1;
  ACanvas->TextRect(RT, Pos.x, Pos.y, sSelected);
      
  // o oldRt V Menu
  RT = oldRT;
  OffsetRect(&RT, 0, FNM.iMenuHeight);
  InflateRect(&RT, 2, 2);
  RT.Bottom = rtActive.Bottom - 2 - FNM.iBorderWidth;
  rtMemo = RT;

  // memo
  Frame3D(ACanvas, RT, FMyColors[ButtonShadow], FMyColors[ButtonHilight], 1);
  Frame3D(ACanvas, RT, FMyColors[ButtonDkShadow], FMyColors[ButtonFace], 1);
  ACanvas->Brush->Color = FMyColors[Window];
  ACanvas->FillRect(RT);

  ACanvas->Font->Name = "System";
  ACanvas->Font->Size = 12;
  ACanvas->Font->Color = clBlack;
  //    TextOut(RT.Left + 2, RT.Top + 2, "r");
  ACanvas->TextOut(RT.Left + 2, RT.Top + 2, sWindowText);

  // scroll box
  RT.Left = RT.Right - FNM.iScrollWidth;
  ACanvas->Brush->Color = FMyColors[ScrollBar];
  ACanvas->FillRect(RT);
      
  RT.Bottom = RT.Top + FNM.iScrollHeight;
  DrawScrollButton(ACanvas, RT, 0); // VW
      
  RT.Bottom = rtMemo.Bottom;
  RT.Top = RT.Bottom - FNM.iScrollHeight;
  DrawScrollButton(ACanvas, RT, 1); // VU
  // }le message box
  RT.Top = rtMemo.Top + (rtMemo.Bottom - rtMemo.Top) * 1 / 3;
  RT.Left = rtMemo.Left + 3;
  RT.Right = rtMemo.Left + (rtMemo.Right - rtMemo.Left) * 3 / 4;
  RT.Bottom = ACanvas->ClipRect.Bottom - 20; // messagebox ݶZe 20 points
      
  // Frame3D |۰ InflateRect...
  rtMsgBox = RT;
  Frame3D(ACanvas, RT, FMyColors[ButtonFace], FMyColors[ButtonDkShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[ButtonHilight], FMyColors[ButtonShadow], 1);
      
  // border
  ACanvas->Brush->Color = FMyColors[ButtonFace];
  ACanvas->FillRect(RT);

  // caption
  InflateRect(&RT, - 1, - 1);
  RT.Bottom = RT.Top + FNM.iCaptionHeight;
  oldRT = RT;
  ACanvas->Brush->Color = FMyColors[ActiveTitle];
      
      
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfCaptionFont]);
  ACanvas->Font->Color = FMyColors[TitleText];
  Pos.x = RT.Left + 2; // LOGFONT
  Pos.y = RT.Top + (RT.Bottom - RT.Top - ACanvas->TextHeight(sMessageBox)) / 2;
  //    TextRect(RT, Pos.x, Pos.y, "T");
  ACanvas->TextRect(RT, Pos.x, Pos.y, sMessageBox);
      
  // eees
  btnSize = FNM.iCaptionWidth - 2;
  InflateRect(&RT, - 2, - 2);
      
  RT.Left = RT.Right - btnSize;
  DrawBorderIcon(ACanvas, RT, 3); // 

  // NONCLIENTMETRICS
  // e message box Tr
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfMessageFont]);
  ACanvas->Font->Color = FMyColors[WindowTextColor];
      
  RT = oldRT;
  Pos.x = RT.Left + 4;
  Pos.y = RT.Top + FNM.iCaptionHeight + 4;
  ACanvas->TextOut(Pos.x, Pos.y, sMessageText);
  RT.Left = Pos.x;
  RT.Top = Pos.y;
  RT.Right = Pos.x + ACanvas->TextWidth(sMessageText);
  RT.Bottom = Pos.y + ACanvas->TextHeight(sMessageText);
      
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfCaptionFont]);
  ACanvas->Font->Color = FMyColors[ButtonText];
  Pos.x = ACanvas->TextWidth(AnsiString(sOK) + "123456");
  Pos.y = ACanvas->TextHeight(sOK) * 3 / 2;
      
  // MessageBox  button
  RT.Left = rtMsgBox.Left + ((rtMsgBox.Right - rtMsgBox.Left) - Pos.x) / 2;
  RT.Right = RT.Left + Pos.x;
  RT.Bottom = rtMsgBox.Bottom - 10;
  RT.Top = RT.Bottom - Pos.y;
      
  Frame3D(ACanvas, RT, FMyColors[ButtonHilight], FMyColors[ButtonDkShadow], 1);
  Frame3D(ACanvas, RT, FMyColors[ButtonFace], FMyColors[ButtonShadow], 1);
  ACanvas->Brush->Color = FMyColors[ButtonFace];
      
  Pos.x = RT.Left + (Pos.x - ACanvas->TextWidth(sOK)) / 2 - 2; // LOGFONT
  Pos.y = RT.Top + (RT.Bottom - RT.Top - ACanvas->TextHeight(sOK)) / 2;
  ACanvas->TextRect(RT, Pos.x, Pos.y, sOK);
      
  // NONCLIENTMETRICS
  // Hint eb Button kU
  // ثe RT O button
  ACanvas->Brush->Color = FMyColors[InfoWindow];
  ACanvas->Font->Color = FMyColors[InfoText];
  ACanvas->Font->Handle = CreateFontIndirect(&FMyFonts[mfMessageFont]);
      
  RT.Left = (RT.Left + RT.Right) / 2;
  RT.Top = RT.Bottom + 5;
  RT.Right = RT.Left + ACanvas->TextWidth(sHint) + ACanvas->TextWidth("12");
  RT.Bottom = RT.Top + ACanvas->TextHeight(sHint);
  ACanvas->TextRect(RT, RT.Left + ACanvas->TextWidth("1"), RT.Top, sHint);
}

void __fastcall TMainForm::FormCreate(TObject *Sender)
{
  PageControl->ActivePage = tabPreview;
  LoadDesc();

  FWallPaperBits = new Graphics::TBitmap;

  LoadCurrentSettings();
  UpdateCaption();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxColorsDrawItem(TWinControl *Control,
      int Index, TRect &Rect, TOwnerDrawState State)
{
  TCanvas* Canvas = lbxColors->Canvas;

  Canvas->FillRect(Rect);
  DrawText(Canvas->Handle, lbxColors->Items->Strings[Index].c_str(),
    lbxColors->Items->Strings[Index].Length(), &Rect, DT_VCENTER | DT_SINGLELINE);

  if (FMyColors[TMyColor(Index)] != -1) {
    Canvas->Brush->Style = bsSolid;
    Canvas->Brush->Color = FMyColors[TMyColor(Index)];
    Rect.Left = Rect.Right - (Rect.Bottom - Rect.Top);
    Canvas->FillRect(Rect);
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxCursorsDrawItem(TWinControl *Control,
      int Index, TRect &Rect, TOwnerDrawState State)
{
  TCanvas* Canvas = lbxCursors->Canvas;

  Canvas->FillRect(Rect);
  DrawText(Canvas->Handle, lbxCursors->Items->Strings[Index].c_str(),
    lbxCursors->Items->Strings[Index].Length(), &Rect, DT_VCENTER | DT_SINGLELINE);

  DrawIcon(Canvas->Handle, Rect.Right - 40, Rect.Top, FMyCursorIcons[TMyCursor(Index)]);
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::lbxFontsDrawItem(TWinControl *Control,
      int Index, TRect &Rect, TOwnerDrawState State)
{
  TCanvas* Canvas = lbxFonts->Canvas;

  Canvas->FillRect(Rect);
  DrawText(Canvas->Handle, lbxFonts->Items->Strings[Index].c_str(),
    lbxFonts->Items->Strings[Index].Length(), &Rect, DT_VCENTER | DT_SINGLELINE);

  Canvas->Font->Handle = CreateFontIndirect(&FMyFonts[TMyFont(Index)]);

  AnsiString S = FontToString(Canvas->Font, false);
  int H = Rect.Bottom - Rect.Top;
  int L = Rect.Right - Canvas->TextWidth(S) - 4;
  if (L < 150) L = 150;
  Canvas->TextOut(L, Rect.Top + (H - Canvas->TextHeight(S)) / 2, S);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxDesktopDrawItem(TWinControl *Control,
      int Index, TRect &Rect, TOwnerDrawState State)
{
  TCanvas* Canvas = lbxDesktop->Canvas;
  Canvas->FillRect(Rect);
  DrawText(Canvas->Handle, lbxDesktop->Items->Strings[Index].c_str(),
    lbxDesktop->Items->Strings[Index].Length(), &Rect, DT_VCENTER | DT_SINGLELINE);

  TRect R;
  switch (Index) {
    case 0: case 1: case 2: case 3:
      DrawIcon(Canvas->Handle, Rect.Right - 40, Rect.Top, FMyIconIcons[TMyIcon(Index)]);
      break;

    case 4: // wallpaper
      R = Rect;
      R.Right = R.Right - 8;
      R.Left = R.Right - 32;
      R.Bottom = R.Top + 32;
      Canvas->StretchDraw(R, FWallPaperBits);
      break;
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxColorsClick(TObject *Sender)
{
  if (FMyColors[TMyColor(lbxColors->ItemIndex)] != -1) {
    pnlColor->Color = FMyColors[TMyColor(lbxColors->ItemIndex)];
    lblRValue->Caption = "j: " + IntToStr(GetRValue(pnlColor->Color));
    lblGValue->Caption = "j: " + IntToStr(GetGValue(pnlColor->Color));
    lblBValue->Caption = "Ŧj: " + IntToStr(GetBValue(pnlColor->Color));
  } else {
    pnlColor->Color = clBtnFace;
    lblRValue->Caption = "";
    lblGValue->Caption = "";
    lblBValue->Caption = "";
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxCursorsClick(TObject *Sender)
{
  imgCursor->Picture->Icon->ReleaseHandle();
  InvalidateRect(pnlCursor->Handle, NULL, true);
  imgCursor->Picture->Icon->Handle = FMyCursorIcons[TMyCursor(lbxCursors->ItemIndex)];

  txtCursorFile->Text = FMyCursors[(TMyCursor)lbxCursors->ItemIndex];
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxSoundsClick(TObject *Sender)
{
  txtSoundFile->Text = FMySounds[TMySound(lbxSounds->ItemIndex)];
  EnableControl(pnlPlaySound,
    /*auxGetNumDevs() != 0 &&*/ FileExists(FMySounds[TMySound(lbxSounds->ItemIndex)]));
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::lbxFontsClick(TObject *Sender)
{
  TLogFont F = FMyFonts[TMyFont(lbxFonts->ItemIndex)];
  lblFontName->Caption = AnsiString("W: ") + F.lfFaceName;
  lblFontSize->Caption = AnsiString("jp: ") + IntToStr(-MulDiv(F.lfHeight, 72, Screen->PixelsPerInch));

  AnsiString S = "˦: ";
  if (F.lfWeight >= 700) S = S + "";
  if (F.lfItalic != 0) S = S + "";
  if (F.lfUnderline != 0) S = S + "u";

  lblFontStyle->Caption = S;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::btnPlaySoundClick(TObject *Sender)
{
  switch (dynamic_cast<TComponent*>(Sender)->Tag) {
    case 0: sndPlaySound(txtSoundFile->Text.c_str(), SND_ASYNC); break;
    case 1: sndPlaySound(NULL, SND_ASYNC); break;
  }
}
//---------------------------------------------------------------------------

typedef bool TPatternArray[8][8];

bool ParsePatternString(AnsiString sPattern, TPatternArray& aPattern)
{
  int x, y, i, val;

  y = 0;
  while (sPattern.Length() > 0) {
    if ((i = sPattern.Pos(" ")) != 0) {
      val = StrToIntDef(sPattern.SubString(1, i - 1), 0);
      sPattern.Delete(1, i);
    } else {
      val = StrToIntDef(sPattern, 0);
      sPattern = "";
    }

    x = 7;
    while (x >= 0) {
      aPattern[y][x] = val & 1;
      val >>= 1;
      x--;
    }

    y++;
  }
  return (y == 8); // successfully parsed
}

void __fastcall TMainForm::ShowPattern()
{
  /* should be "(None)", but Windows NT 4.0 Traditional Chinese
    version translated it to "(L)". */
  const char* NON_PATTERN_STR = "(None)";

  if (FPattern == NON_PATTERN_STR) return;

  TPatternArray aPattern;

  // decomposite the pattern string
  if (!ParsePatternString(FPattern, aPattern)) return;

  // draw each cell base on aPattern boolean array
  // calculate the width of each cell and adjust total width
  int w = imgIcon->ClientWidth / 8;
  imgIcon->ClientWidth = w * 8;

  // calculate the height of each cell and adjust total height
  int h = imgIcon->ClientHeight / 8;
  imgIcon->ClientHeight = h * 8;

  imgIcon->Canvas->Brush->Style = bsSolid;
  imgIcon->Canvas->Brush->Color = clBackground;
  /* first fill all the canvas with clBackground color
    (corresponding COLOR_BACKGROUND in WINDOWS->PAS) */
  imgIcon->Canvas->FillRect(imgIcon->Canvas->ClipRect);

  // then draw the coloring cell with black color
  imgIcon->Canvas->Brush->Color = clBlack;

  for (int y = 0; y < 8; y++)
    for (int x = 0; x < 8; x++)
      if (aPattern[y][x])
        imgIcon->Canvas->FillRect(Rect(x * w, y * h, (x + 1) * w, (y + 1) * h));
}

void __fastcall TMainForm::lbxDesktopClick(TObject *Sender)
{
  imgIcon->Picture->Icon->ReleaseHandle();
  InvalidateRect(pnlIcon->Handle, NULL, true);
  SendMessage(GetWindow(pnlIcon->Handle, GW_CHILD), WM_CLOSE, 0, 0);
  btnPreview->Enabled = false;

  if (lbxDesktop->ItemIndex < 7) nbkDesktop->PageIndex = 0;

  int ItemIndex = lbxDesktop->ItemIndex;
  TRect R;
  
  switch (ItemIndex) {
    case 0: case 1: case 2: case 3: // desktop icons
      txtDesktopFile->Text = TruncateTrailIfNotDLL(FMyIcons[TMyIcon(ItemIndex)]);
      imgIcon->Picture->Icon->Handle = FMyIconIcons[TMyIcon(ItemIndex)];
      break;

    case 4: // Wallpaper
      txtDesktopFile->Text = FWallPaper;

      if (FWallPaper != "") { // valid wallpaper
        R = Rect(0, 0, imgIcon->ClientWidth, imgIcon->ClientHeight);
        imgIcon->Picture->Bitmap->Width = imgIcon->ClientWidth;
        imgIcon->Picture->Bitmap->Height = imgIcon->ClientHeight;
        imgIcon->Canvas->StretchDraw(R, FWallPaperBits);
      }
      break;

    case 5: // pattern
      txtDesktopFile->Text = "";

      ShowPattern();
      break;

    case 6: // screen saver
      txtDesktopFile->Text = FScreenSaver;

      if (FScreenSaver != "") {
        WinExec(Format("%s -p %d", OPENARRAY(TVarRec, (FScreenSaver, (int)pnlIcon->Handle))).c_str(), SW_SHOW);
        btnPreview->Enabled = true;
      }
      break;

    case 7: // window metrics
      nbkDesktop->PageIndex = 1;
      Application->ProcessMessages();

      lblBorderWidth->Caption = AnsiString("ؼe: ") + IntToStr(FNM.iBorderWidth);
      lblTitleWrap->Caption = AnsiString("ϥܤr: ") + IntToStr(FIM.iTitleWrap);
      cbxMetricsKind->ItemIndex = 0;
      cbxMetricsKind->OnChange(cbxMetricsKind);
      break;

    default:
      txtDesktopFile->Text = "";
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::cbxMetricsKindChange(TObject *Sender)
{

  int H, V;

  switch (cbxMetricsKind->ItemIndex) {
    case 0:
      H = FNM.iMenuWidth;
      V = FNM.iMenuHeight;
      break;

    case 1:
      H = FNM.iCaptionWidth;
      V = FNM.iCaptionHeight;
      break;

    case 2:
      H = FNM.iScrollWidth;
      V = FNM.iScrollHeight;
      break;

    case 3:
      H = FIM.iHorzSpacing;
      V = FIM.iVertSpacing;
      break;

    case 4:
      H = FNM.iSmCaptionWidth;
      V = FNM.iSmCaptionHeight;
      break;
  }

  lblHorizontal->Caption = AnsiString(": ") + IntToStr(H);
  lblVertical->Caption = AnsiString(": ") + IntToStr(V);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
  delete FWallPaperBits;        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuApplyClick(TObject *Sender)
{
  if (FThemeName == "") return;

  ApplyTheme();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::btnPreviewClick(TObject *Sender)
{
  if (FScreenSaver != "")
    WinExec(Format("%s -s", OPENARRAY(TVarRec, (FScreenSaver))).c_str(), SW_SHOW);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuCurrentClick(TObject *Sender)
{
  LoadCurrentSettings();
  UpdateCaption();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuExitClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuAboutClick(TObject *Sender)
{
  TAboutBox* box = new TAboutBox(this);
  try {
    box->ShowModal();
  } __finally {
    delete box;
  }
}
//---------------------------------------------------------------------------

