unit xGraphics;

interface

uses Windows, Messages, SysUtils, Forms, Controls, Graphics, ShellAPI, Registry, Math, Dialogs;

function ColorToRGBString(AColor: TColor): string;
function RGBStringToColor(RGBStr: string): TColor;

function MyLoadIcon(const Filename: string; DefaultIcon: THandle = 0): THandle;
procedure GetIconBits(IconHandle: HICON; Bits: TBitmap);

function CursorToIcon(const hc: THandle; bDestroyCursor: Boolean = False): THandle;

function ColorNumToStr(Num: Integer): string;
function ColorDiff(C1, C2: TColor): Integer;

procedure ReplaceBitmapColor(Bits: TBitmap; FromColor, ToColor: COLORREF);

function ColorToHexStr(C: TColor): string;
function HexStrToColor(S: string): TColor;

function ContrastColor(const mc: TColor): TColor;
function OppositeColor(C: TColor): TColor;

implementation

uses xUtils, xStrings, xFiles;

function ColorToRGBString(AColor: TColor): string;
begin
  Result := Format('%d %d %d',[GetRValue(AColor), GetGValue(AColor), GetBValue(AColor)]);
end;

function RGBStringToColor(RGBStr: string): TColor;
var
  I      : Integer;
  R, G, B: Byte;   
begin
  Result := $00000000;
  RGBStr := Trim(RGBStr);
  if CountWords(RGBStr) <> 3 then Exit;
  
  // Red Value
  I := Pos(' ', RGBStr);
  R := StrToIntDef(Copy(RGBStr, 1, I - 1), 0);
  Delete(RGBStr, 1, I);
  
  // Green Value
  I := Pos(' ', RGBStr);
  G := StrToIntDef(Copy(RGBStr, 1, I - 1), 0);
  Delete(RGBStr, 1, I);
  
  // Blue
  B := StrToIntDef(RGBStr, 0);
  
  Result := RGB(R, G, B);
end;

function MyLoadIcon(const Filename: string; DefaultIcon: THandle = 0): THandle;
var
  IconNo: Integer;
  S, Ext: string; 
begin
  Result := 0;
  
  S := AnsiUpperCase(Filename);
  IconNo := TruncateTrailNumber(S);
  
  Ext := ExtractFileExt(S);
  if Ext = '.ICO' then
    Result := LoadImage(HInstance, PChar(S), IMAGE_ICON, 32, 32,
    LR_LOADFROMFILE)
  else if ((Ext = '.ICL') or (Ext = '.DLL') or (Ext = '.EXE'))
    and (IconNo <> -1) then
    Result := ExtractIcon(HInstance, PChar(S), IconNo);
  
  // pG, N default icon N..
  if Result = 0 then
    Result := DefaultIcon;
  // pG default icon ]sb, Nε{ icon N
  if Result = 0 then
    Result := Application.Icon.Handle;
end;

// NйϮର HICON HK
function CursorToIcon(const hc: THandle; bDestroyCursor: Boolean = False): THandle;
var
  ii: TIconInfo;
begin
  if hc <> 0 then
  begin
    GetIconInfo(hc, ii);
    Result := CreateIconIndirect(ii);
    if bDestroyCursor then DestroyCursor(hc);
  end else
    Result := 0;
end;

function ColorNumToStr(Num: Integer): string;
begin
  case Num of
    16: Result := '4 bit';
    256: Result := '8 bit';
    32768: Result := '15 bit';
    65536: Result := '16 bit';
    16777216: Result := '24 bit';
  else Result := IntToStr(Num) + 'colors';
  end;
end;

function ColorDiff(C1, C2: TColor): Integer;
begin
  Result := Abs(GetRValue(C1) - GetRValue(C2)) +
    Abs(GetGValue(C1) - GetGValue(C2)) +
    Abs(GetBValue(C1) - GetBValue(C2));
end;

procedure ReplaceBitmapColor(Bits: TBitmap; FromColor, ToColor: COLORREF);
var
  Y, X: Integer;
  P   : PDWORD; 
begin
  Bits.PixelFormat := pf32bit;
  
  with Bits do
    for Y := 0 to Height - 1 do
    begin
      P := Bits.ScanLine[Y];
      for X := 0 to Width - 1 do
      begin
        if P^ = FromColor then P^ := ToColor;
        Inc(P);
      end;
    end;
end;

function ColorToHexStr(C: TColor): string;
begin
  Result := IntToHex(GetRValue(C), 2) +
    IntToHex(GetGValue(C), 2) +
    IntToHex(GetBValue(C), 2);
end;

function HexStrToColor(S: string): TColor;
var
  R, G, B: Byte;
begin
  Result := clWhite; // w]C
  
  if S = '' then Exit;
  // maybe #RRGGBB either RRGGBB
  if S[1] = '#' then Delete(S, 1, 1);
  
  if (Length(S) <> 6) then Exit;
  
  try
    R := HexToInt(Copy(S, 1, 2));
    G := HexToInt(Copy(S, 3, 2));
    B := HexToInt(Copy(S, 5, 2));
  except
    Exit;
  end;
  
  Result := RGB(R, G, B);
end;

function ContrastColor(const mc: TColor): TColor;
var
  R, G, B: Byte;
begin
  R := GetRValue(mc);
  G := GetGValue(mc);
  B := GetBValue(mc);
  
  if (R + G + B) div 3 > 128 then Result := clBlack else Result := clWhite;
end;

function OppositeColor(C: TColor): TColor;
var
  R, G, B: Byte;
begin
  R := GetRValue(C);
  G := GetGValue(C);
  B := GetBValue(C);
  
  Result := RGB(255 - R, 255 - G, 255 - B);
end;

procedure GetIconBits(IconHandle: HICON; Bits: TBitmap);
var
  Icon: TIcon;
begin
  Bits.Width := SM_CXICON;
  Bits.Height := SM_CYICON;

  Icon := TIcon.Create;
  with Icon do
    try
      Handle := IconHandle;
      Bits.Canvas.Draw(0, 0, Icon);
    finally
      Free;
    end;
end;

end.

