unit xUtils;

interface

uses Windows, Messages, Classes, SysUtils, Forms, Controls;

procedure MsgBox(const Msg: string);
procedure ErrBox(const Msg: string);
function YesNoBox(const Msg: string; DefButton: DWORD = MB_DEFBUTTON1): Boolean;
function YesNoCancelBox(const Msg: string): Integer;

procedure DoBusy(Busy: Boolean);

procedure ShowLastError(const Msg: string = 'API Error');
procedure RaiseLastError(const Msg: string = 'API Error');

procedure FreeStringsObjects(SL: TStrings);
procedure FreeStringsPChars(SL: TStrings);

procedure FreeListPointers(List: TList);

function HexToInt(Hex: string): Integer; //HEX ഫ Integer

function GetShiftPressed: Boolean;
function GetControlPressed: Boolean;

type
  TSortCompareFunc = function (P1, P2: Pointer): Integer;
  
procedure InsertionSort(P: Pointer; Num, Size: Integer; SortFunc: TSortCompareFunc);

function IntegerSortCompareFunc(P1, P2: Pointer): Integer;

implementation

procedure MsgBox(const Msg: string);
begin
  Application.MessageBox(PChar(Msg), PChar(Application.Title), MB_ICONINFORMATION);
end;

procedure ErrBox(const Msg: string);
begin
  Application.MessageBox(PChar(Msg), PChar(Application.Title), MB_ICONERROR);
end;

function YesNoBox(const Msg: string; DefButton: DWORD = MB_DEFBUTTON1): Boolean;
begin
  Result := Application.MessageBox(PChar(Msg), PChar(Application.Title), MB_ICONQUESTION or
    MB_YESNO or DefButton) = IDYES;
end;

function YesNoCancelBox(const Msg: string): Integer;
begin
  Result := Application.MessageBox(PChar(Msg),
    PChar(Application.Title), MB_ICONQUESTION or MB_YESNOCANCEL or MB_DEFBUTTON1)
end;

procedure DoBusy(Busy: Boolean);
const
  Times: Integer = 0;
begin
  if Busy then
  begin
    Inc(Times);
    if Times = 1 then Screen.Cursor := crHourGlass;
  end else
  begin
    dec(Times);
    if Times = 0 then Screen.Cursor := crDefault;
  end;
end;

function GetLastErrorStr: string;
var
  Buf: PChar;
begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM,
    nil, GetLastError, LANG_USER_DEFAULT, @Buf, 0, nil);
  try
    Result := StrPas(Buf);
  finally
    LocalFree(HLOCAL(Buf));
  end;
end;

procedure ShowLastError(const Msg: string = 'API Error');
begin
  MsgBox(Msg + ': ' + GetLastErrorStr);
end;

procedure RaiseLastError(const Msg: string = 'API Error');
var
  S: string;
begin
  S := GetLastErrorStr;
  if S = '' then S := IntToStr(GetLastError);
  raise Exception.Create(Msg + ': ' + S);
end;

procedure FreeStringsObjects(SL: TStrings);
var
  I: Integer;
begin
  for I := 0 to SL.Count - 1 do
    if Assigned(SL.Objects[I]) then
    begin
      Dispose(Pointer(SL.Objects[I]));
      SL.Objects[I] := nil;
    end;
end;

procedure FreeStringsPChars(SL: TStrings);
var
  I: Integer;
begin
  for I := 0 to SL.Count - 1 do
    if Assigned(SL.Objects[I]) then
    begin
      StrDispose(PChar(SL.Objects[I]));
      SL.Objects[I] := nil;
    end;
end;

procedure InsertionSort(P: Pointer; Num, Size: Integer; SortFunc: TSortCompareFunc);
var
  I, J, Min_No: Integer;
  T           : PChar;  
begin
  GetMem(T, Size);
  try
    // insertion sort (stable sort)
    for I := 0 to Num - 1 do
    begin
      Min_No := I;
      for J := I + 1 to Num - 1 do
        if SortFunc(Pointer(Integer(P) + Size * J), Pointer(Integer(P) + Size * Min_No)) < 0 then
          Min_No := J;
      
      if I <> Min_No then
      begin
        Move(Pointer(Integer(P) + Size * I)^, T^, Size);
        Move(Pointer(Integer(P) + Size * Min_No)^, Pointer(Integer(P) + Size * I)^, Size);
        Move(T^, Pointer(Integer(P) + Size * Min_No)^, Size);
      end;
    end;
  finally
    FreeMem(T);
  end;
end;

function IntegerSortCompareFunc(P1, P2: Pointer): Integer;
type
  PInteger = ^Integer;
var
  I1, I2: Integer;
begin
  I1 := PInteger(P1)^;
  I2 := PInteger(P2)^;
  
  if I1 < I2 then Result := -1
  else if I1 > I2 then Result := 1
  else Result := 0;
end;

// = [ HEX ഫ Integer ] ==============================
function HexToInt(Hex: string): Integer; //HEX ഫ Integer
const
  Stext = '123456789ABCDEF';
  {var
  I, LenNo, Sum: Integer;      }
var
  I, Power: Integer;
begin
  //  ShowMessage('HexToInt parameter: ' + Hex);
  {
   // շlg
  Sum := 0;
  LenNo := Length(UpperCase(Hex)); //o
  for I := LenNo - 1 downto 0 do
    Sum := Sum + IntPower(16, I) * Pos(Hex[LenNo - I], Stext);
  
  Result := Sum;
  }
  
  Hex := UpperCase(Hex);
  
  Result := 0;
  Power := 1;
  for I := Length(Hex) downto 1 do
  begin
    Result := Result + Pos(Hex[I], Stext) * Power;
    Power := Power * 16;
  end;
end;

function GetShiftPressed: Boolean;
begin
  Result := ((GetKeyState(VK_LSHIFT) and $8000) <> 0) or ((GetKeyState(VK_RSHIFT) and $8000) <> 0);
end;

function GetControlPressed: Boolean;
begin
  Result := ((GetKeyState(VK_LCONTROL) and $8000) <> 0) or ((GetKeyState(VK_RCONTROL) and $8000) <> 0);
end;

procedure FreeListPointers(List: TList);
var
  I: Integer;
begin
  for I := 0 to List.Count - 1 do
    if Assigned(List[I]) then
    begin
      Dispose(List[I]);
      List[i] := nil;
    end;
end;

end.

